from _decimal import Decimal
from typing import Any, Dict, List

import openpyxl
from openpyxl.worksheet.worksheet import Worksheet

from sirius.exceptions import SDKClientException


def _get_cell_value_from_cell(cell_value: Any) -> Any:
    return Decimal(str(round(cell_value, 6))) if isinstance(cell_value, (int, float)) and not isinstance(cell_value, bool) else cell_value


def _get_worksheet(file_path: str, sheet_name: str) -> Worksheet:
    try:
        workbook = openpyxl.load_workbook(filename=file_path, data_only=True)
    except FileNotFoundError:
        raise SDKClientException(f"Excel file not found in: {file_path}")

    if sheet_name not in workbook:
        raise SDKClientException(f"Excel sheet not found\n: "
                                 f"File path: {file_path}\n"
                                 f"Sheet name: {sheet_name}")
    return workbook[sheet_name]


def get_excel_data(file_path: str, sheet_name: str) -> List[Dict[Any, Any]]:
    excel_data_list: List[Dict[Any, Any]] = []
    headers: List[Any] = []

    row_number: int = 0
    for row in _get_worksheet(file_path, sheet_name):
        if row_number == 0:
            headers = list(map(lambda c: c.value, filter(lambda c: c.value is not None, row)))

        else:
            excel_data: Dict[Any, Any] = {}
            cell_number: int = 0
            for cell in row:
                excel_data[headers[cell_number]] = _get_cell_value_from_cell(cell.value)
                cell_number = cell_number + 1

            excel_data_list.append(excel_data)

        row_number = row_number + 1

    return excel_data_list


def get_key_value_pair(file_path: str, sheet_name: str) -> Dict[Any, Any]:
    key_value_pair: Dict[Any, Any] = {}
    for row in _get_worksheet(file_path, sheet_name):
        key_value_pair[str(row[0].value)] = _get_cell_value_from_cell(row[1].value)

    return key_value_pair



import asyncio
import base64
import datetime
import inspect
import io
import multiprocessing
import os
import secrets
import socket
import string
import tempfile
import threading
from _decimal import Decimal
from concurrent.futures import ProcessPoolExecutor
from enum import Enum
from typing import Callable, Any, Dict, List

import pytz
import qrcode
import requests
from pydantic import BaseModel, ConfigDict
from qrcode.image.pil import PilImage

from sirius.constants import EnvironmentVariable, EnvironmentSecret
from sirius.exceptions import ApplicationException, SDKClientException, OperationNotSupportedException


class Environment(Enum):
    Production: str = "Production"
    Test: str = "Test"
    Development: str = "Development"
    CI_CD_PIPELINE: str = "CI/CD Pipeline"


class Currency(Enum):
    AED: str = "AED"
    AUD: str = "AUD"
    BDT: str = "BDT"
    BGN: str = "BGN"
    CAD: str = "CAD"
    CHF: str = "CHF"
    CLP: str = "CLP"
    CNY: str = "CNY"
    CRC: str = "CRC"
    CZK: str = "CZK"
    DKK: str = "DKK"
    EGP: str = "EGP"
    EUR: str = "EUR"
    GBP: str = "GBP"
    GEL: str = "GEL"
    HKD: str = "HKD"
    HUF: str = "HUF"
    IDR: str = "IDR"
    ILS: str = "ILS"
    INR: str = "INR"
    JPY: str = "JPY"
    KES: str = "KES"
    KRW: str = "KRW"
    LKR: str = "LKR"
    MAD: str = "MAD"
    MXN: str = "MXN"
    MYR: str = "MYR"
    NGN: str = "NGN"
    NOK: str = "NOK"
    NPR: str = "NPR"
    NZD: str = "NZD"
    PHP: str = "PHP"
    PKR: str = "PKR"
    PLN: str = "PLN"
    RON: str = "RON"
    SEK: str = "SEK"
    SGD: str = "SGD"
    THB: str = "THB"
    TRY: str = "TRY"
    TZS: str = "TZS"
    UAH: str = "UAH"
    UGX: str = "UGX"
    USD: str = "USD"
    UYU: str = "UYU"
    VND: str = "VND"
    XOF: str = "XOF"
    ZAR: str = "ZAR"


class DataClass(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


def get_environmental_variable(environmental_variable: EnvironmentVariable | str) -> str:
    environmental_variable_key: str = environmental_variable.value if isinstance(environmental_variable,
                                                                                 EnvironmentVariable) else environmental_variable
    value: str | None = os.getenv(environmental_variable_key)
    if value is None:
        raise ApplicationException(f"Environment variable with the key is not available: {environmental_variable_key}")

    return value


def get_environmental_secret(environmental_secret: EnvironmentSecret | str) -> str:
    from sirius.azure.key_vault import AzureKeyVault
    environmental_secret_key: str = environmental_secret.value if isinstance(environmental_secret,
                                                                             EnvironmentSecret) else environmental_secret
    return AzureKeyVault.get(environmental_secret_key)


def get_environment() -> Environment:
    environment: str | None = os.getenv(EnvironmentVariable.ENVIRONMENT.value)
    try:
        return Environment.Development if environment is None else Environment(environment)
    except ValueError:
        raise ApplicationException(f"Invalid environment variable setup: {environment}")


def is_production_environment() -> bool:
    return Environment.Production == get_environment()


def is_test_environment() -> bool:
    return Environment.Test == get_environment()


# TODO: Create redundancy check (if a test/production environment is identified as development, no authentication is done)
def is_development_environment() -> bool:
    return Environment.Development == get_environment() or is_ci_cd_pipeline_environment()


def is_ci_cd_pipeline_environment() -> bool:
    return Environment.CI_CD_PIPELINE == get_environment()


def get_application_name() -> str:
    return get_environmental_secret(EnvironmentSecret.APPLICATION_NAME)


def threaded(func: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        thread: threading.Thread = threading.Thread(target=func, args=args, kwargs=kwargs)
        thread.start()
        return thread

    return wrapper


def wait_for_all_coroutines(func: Callable) -> Callable:
    if not inspect.iscoroutinefunction(func):
        raise SDKClientException(
            f"Synchronous method used to in asynchronous context: {func.__module__}.{func.__name__}")

    # TODO: Manage Exceptions
    async def wrapper(*args: Any, **kwargs: Any) -> None:
        asyncio.create_task(func(*args, **kwargs))
        await asyncio.wait(  # type: ignore[type-var]
            set(filter(lambda t: ("wait_for_all_coroutines" not in t.get_coro().__qualname__), asyncio.all_tasks())),  # type: ignore[arg-type,union-attr,attr-defined]
            timeout=600)
        return None

    return wrapper


def is_dict_include_another_dict(one_dict: Dict[Any, Any], another_dict: Dict[Any, Any]) -> bool:
    if not all(key in one_dict for key in another_dict):
        return False

    for key, value in one_dict.items():
        if another_dict[key] != value:
            return False

    return True


def get_decimal_str(decimal: Decimal) -> str:
    return "{:,.2f}".format(float(decimal))


def only_in_dev(func: Callable) -> Callable:
    def wrapper(*args: Any, **kwargs: Any) -> threading.Thread:
        if not is_development_environment():
            raise OperationNotSupportedException("Operation is only permitted in the dev environment")
        return func(*args, **kwargs)

    return wrapper


def get_servers_fqdn() -> str:
    return socket.getfqdn()


def get_timestamp_from_string(timestamp_string: str, timezone_string: str | None = None) -> datetime.datetime:
    timestamp: datetime.datetime = datetime.datetime.strptime(timestamp_string, "%Y-%m-%dT%H:%M:%SZ")
    return timestamp if timezone_string is None else timestamp.replace(tzinfo=pytz.timezone(timezone_string))


def get_new_temp_file_path() -> str:
    return tempfile.NamedTemporaryFile(delete=False).name


def download_file_from_url(url: str) -> str:
    temp_file_path: str = get_new_temp_file_path()
    open(temp_file_path, "wb").write(requests.get(url).content)
    return temp_file_path


def get_unique_id(length: int = 16) -> str:
    raw_id: str = "".join(
        secrets.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits) for _ in range(length))
    return "-".join([raw_id[i:i + 4] for i in range(0, len(raw_id), 4)])


def run_function_using_multiple_processes(func: Callable, argument_list: List[Any],
                                          maximum_number_of_threads: int | None = None) -> List[Any | None]:
    maximum_number_of_threads = multiprocessing.cpu_count() + 1 if maximum_number_of_threads is None else maximum_number_of_threads
    maximum_number_of_threads = min(maximum_number_of_threads, len(argument_list))

    with ProcessPoolExecutor(max_workers=maximum_number_of_threads) as executor:
        return list(executor.map(func, argument_list))


def get_qr_code(data_str: str) -> str:
    buffered: io.BytesIO = io.BytesIO()
    qr_code: PilImage = qrcode.make(data_str)

    qr_code.save(buffered, format="PNG")
    return base64.b64encode(buffered.getvalue()).decode()


def get_base64_string_from_bytes(data: bytes) -> str:
    return base64.b64encode(data).decode("utf-8")


def get_bytes_from_base64_string(base64_string: str) -> bytes:
    return base64.b64decode(base64_string)



from sirius import common

URL: str = "https://api.transferwise.com" if common.is_production_environment() else "https://api.sandbox.transferwise.tech"
ENDPOINT__PROFILE__GET_ALL: str = f"{URL}/v2/profiles"
ENDPOINT__ACCOUNT__GET_ALL__CASH_ACCOUNT: str = f"{URL}/v4/profiles/$profileId/balances?types=STANDARD"
ENDPOINT__ACCOUNT__GET_ALL__RESERVE_ACCOUNT: str = f"{URL}/v4/profiles/$profileId/balances?types=SAVINGS"

ENDPOINT__BALANCE__MOVE_MONEY_BETWEEN_BALANCES: str = f"{URL}/v2/profiles/$profileId/balance-movements"
ENDPOINT__BALANCE__OPEN: str = f"{URL}/v3/profiles/$profileId/balances"
ENDPOINT__BALANCE__CLOSE: str = f"{URL}/v3/profiles/$profileId/balances/$balanceId"
ENDPOINT__BALANCE__GET_TRANSACTIONS: str = f"{URL}/v1/profiles/$profileId/balance-statements/$balanceId/statement.json"
ENDPOINT__TRANSFER__CREATE_THIRD_PARTY_TRANSFER: str = f"{URL}/v1/transfers"
ENDPOINT__TRANSFER__FUND_THIRD_PARTY_TRANSFER: str = f"{URL}/v3/profiles/$profileId/transfers/$transferId/payments"
ENDPOINT__TRANSFER__GET_ALL: str = f"{URL}/v1/transfers??profile=$profileId"
ENDPOINT__DEBIT_CARD__GET_ALL: str = f"{URL}/v3/spend/profiles/$profileId/cards"

ENDPOINT__QUOTE__GET: str = f"{URL}/v3/profiles/$profileId/quotes"
ENDPOINT__INTRA_ACCOUNT_TRANSFER__CREATE: str = f"{URL}/v2/profiles/$profileId/balance-movements"
ENDPOINT__RECIPIENT__GET_ALL: str = f"{URL}/v1/accounts?profile=$profileId"

ENDPOINT__SIMULATION__TOP_UP: str = f"{URL}/v1/simulation/balance/topup"
ENDPOINT__SIMULATION__COMPLETE_TRANSFER: str = f"{URL}/v1/simulation/transfers/$transferId/$status"



import asyncio

from twilio.rest import Client
from twilio.rest.api.v2010.account.message import MessageInstance

from sirius import common
from sirius.constants import EnvironmentSecret

twillo_client: Client | None = None


async def send_message(phone_number_string: str, message: str) -> MessageInstance:
    return await asyncio.get_event_loop().run_in_executor(None, _sync_send_message, phone_number_string, message)


def _sync_send_message(phone_number_string: str, message: str) -> MessageInstance:
    global twillo_client
    twillo_account_sid: str = common.get_environmental_secret(EnvironmentSecret.TWILIO_ACCOUNT_SID)
    twillo_auth_token: str = common.get_environmental_secret(EnvironmentSecret.TWILIO_AUTH_TOKEN)
    twillo_client = Client(twillo_account_sid, twillo_auth_token) if twillo_client is None else twillo_client

    return twillo_client.messages.create(
        from_=f"{common.get_environmental_secret(EnvironmentSecret.TWILIO_SMS_NUMBER)}",
                                         body=message,
                                         to=f"{phone_number_string}")



from enum import Enum


class EnvironmentVariable(Enum):
    ENVIRONMENT: str = "ENVIRONMENT"
    AZURE_KEY_VAULT_URL: str = "AZURE_KEY_VAULT_URL"
    AZURE_CLIENT_ID: str = "AZURE_CLIENT_ID"
    AZURE_CLIENT_SECRET: str = "AZURE_CLIENT_SECRET"
    AZURE_TENANT_ID: str = "AZURE_TENANT_ID"


class EnvironmentSecret(Enum):
    APPLICATION_NAME: str = "APPLICATION-NAME"
    SENTRY_URL: str = "SENTRY-URL"
    MONGO_DB_CONNECTION_STRING: str = "MONGO-DB-CONNECTION-STRING"
    DATABASE_NAME: str = "DATABASE-NAME"
    DISCORD_BOT_TOKEN: str = "DISCORD-BOT-TOKEN"
    WISE_PRIMARY_ACCOUNT_API_KEY: str = "WISE-PRIMARY-ACCOUNT-API-KEY"
    WISE_SECONDARY_ACCOUNT_API_KEY: str = "WISE-SECONDARY-ACCOUNT-API-KEY"
    ENTRA_ID_CLIENT_ID: str = "ENTRA-ID-CLIENT-ID"
    ENTRA_ID_TENANT_ID: str = "ENTRA-ID-TENANT-ID"
    TWILIO_AUTH_TOKEN: str = "TWILIO-AUTH-TOKEN"
    TWILIO_ACCOUNT_SID: str = "TWILIO-ACCOUNT-SID"
    TWILIO_WHATSAPP_NUMBER: str = "TWILIO-WHATSAPP-NUMBER"
    TWILIO_SMS_NUMBER: str = "TWILIO-SMS-NUMBER"



import base64

import pyotp


class TwoFactorAuthentication:

    @staticmethod
    def get_authenticator_uri(username: str, issuer_name: str, hash_str: str | None) -> str:
        hash_str = pyotp.random_base32() if hash_str is None else base64.b32encode(hash_str.encode("UTF-8")).decode("UTF-8")
        return pyotp.totp.TOTP(hash_str).provisioning_uri(name=username, issuer_name=issuer_name)

    @staticmethod
    def get_otp(hash_str: str) -> str:
        return pyotp.TOTP(base64.b32encode(hash_str.encode("UTF-8")).decode("UTF-8")).now()

    @staticmethod
    def is_otp_valid(otp: str, hash_str: str) -> bool:
        return pyotp.TOTP(base64.b32encode(hash_str.encode("UTF-8")).decode("UTF-8")).verify(otp)



import datetime
import json
from typing import Dict, Any, cast, Union, Callable

import jwt
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from fastapi import Request
from jwt import InvalidTokenError, ExpiredSignatureError

from sirius.azure.key_vault import AzureKeyVault
from sirius.common import DataClass
from sirius.exceptions import SDKClientException
from sirius.iam import constants
from sirius.iam.exceptions import InvalidAccessTokenException
from sirius.iam.microsoft_entra_id import MicrosoftIdentity


class Identity(DataClass):
    microsoft_identity: MicrosoftIdentity
    ip_address: str | None = None
    port_number: int | None = None

    @staticmethod
    def get_private_key() -> RSAPrivateKey:
        try:
            private_key_string: str = AzureKeyVault.get(constants.AUTHENTICATION_KEY_NAME)
            return cast(RSAPrivateKey, serialization.load_pem_private_key(private_key_string.encode("utf-8"), None, backend=default_backend()))
        except SDKClientException:
            private_key: RSAPrivateKey = rsa.generate_private_key(public_exponent=65537, key_size=4096, backend=default_backend())
            AzureKeyVault.set(constants.AUTHENTICATION_KEY_NAME, private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption()
            ).decode("utf-8"))

            return private_key

    @staticmethod
    async def get_access_token_remotely(redirect_url: str,
                                        client_ip_address: str | None = None,
                                        client_port_number: int | None = None,
                                        url_shortener_function: Union[Callable, None] = None) -> str:

        microsoft_access_token: str = await MicrosoftIdentity.get_access_token_remotely(redirect_url, url_shortener_function=url_shortener_function)
        microsoft_identity: MicrosoftIdentity = MicrosoftIdentity.get_identity_from_access_token(microsoft_access_token)
        identity: Identity = Identity._get_identity_from_microsoft_identity(microsoft_identity, client_ip_address, client_port_number)
        return Identity.get_access_token_from_identity(identity)

    @staticmethod
    def get_identity_from_request(request: Request) -> "Identity":
        if request.headers.get("Authorization") is None or "Bearer " not in request.headers.get("Authorization") or request.cookies.get("access_token") is None or "Bearer " not in request.cookies["access_token"]:
            raise InvalidAccessTokenException("Invalid Token")

        access_token: str = (request.headers.get("Authorization") if "Authorization" in request.headers else request.cookies.get("Authorization")).replace("Bearer ", "")
        return Identity.get_identity_from_access_token(access_token)

    @staticmethod
    def get_identity_from_access_token(access_token: str) -> "Identity":
        Identity._validate_jwt_token(access_token)
        payload: Dict[str, Any] = jwt.decode(access_token, options={"verify_signature": False})

        return Identity(**payload)

    @staticmethod
    def get_access_token_from_identity(identity: "Identity") -> str:
        payload: Dict[str, Any] = json.loads(identity.model_dump_json())
        payload["exp"] = datetime.datetime.utcnow() + datetime.timedelta(seconds=constants.DEFAULT_ACCESS_TOKEN_VALIDITY_SECONDS)

        return jwt.encode(payload, Identity.get_private_key(), algorithm="RS256")

    @staticmethod
    def _validate_jwt_token(jwt_token: str) -> None:
        try:
            jwt.decode(jwt_token, Identity.get_private_key().public_key(), algorithms=["RS256"])
        except ExpiredSignatureError:
            raise InvalidAccessTokenException("Token has expired")
        except InvalidTokenError:
            raise InvalidAccessTokenException("Token's cryptographic verification failed")

    @staticmethod
    def _get_identity_from_microsoft_identity(microsoft_identity: MicrosoftIdentity, ip_address: str | None = None, port_number: int | None = None) -> "Identity":
        return Identity(microsoft_identity=microsoft_identity, ip_address=ip_address, port_number=port_number)



import datetime
import uuid
from _decimal import Decimal, ROUND_HALF_UP
from enum import Enum, auto
from typing import List, Dict, Any, Union, cast

from pydantic import PrivateAttr, Field

from sirius import common
from sirius.common import DataClass, Currency
from sirius.communication.discord import AortaTextChannels, get_timestamp_string, \
    DiscordDefaults
from sirius.constants import EnvironmentSecret
from sirius.exceptions import OperationNotSupportedException, SDKClientException
from sirius.http_requests import SyncHTTPSession, HTTPResponse
from sirius.http_requests.exceptions import HTTPException
from sirius.wise import constants
from sirius.wise.exceptions import CashAccountNotFoundException, ReserveAccountNotFoundException, \
    RecipientNotFoundException


class WiseAccountType(Enum):
    PRIMARY = auto()
    SECONDARY = auto()


class TransactionType(Enum):
    CARD: str = "CARD"
    CONVERSION: str = "CONVERSION"
    DEPOSIT: str = "DEPOSIT"
    TRANSFER: str = "TRANSFER"
    MONEY_ADDED: str = "MONEY_ADDED"
    UNKNOWN: str = "UNKNOWN"


class WebhookAccountUpdateType(Enum):
    CREDIT: str = "balances#credit"
    UPDATE: str = "balances#update"
    STATE_CHANGE: str = "balances#account-state-change"


class WiseAccount(DataClass):
    type: WiseAccountType
    personal_profile: "PersonalProfile"
    business_profile: "BusinessProfile"
    _http_session: SyncHTTPSession = PrivateAttr()

    @property
    def http_session(self) -> SyncHTTPSession:
        return self._http_session

    def _initialize(self) -> None:
        if (self.personal_profile is None and self.business_profile is not None) or (
                self.personal_profile is not None and self.business_profile is None):
            raise SDKClientException(
                "One profile has been de-initialized; profile attributes should never be de-initialized in the code")

        if self.personal_profile is None or self.business_profile is None:
            profile_list: List[Profile] = Profile.get_all(self)
            self.personal_profile = cast(PersonalProfile,
                                         next(filter(lambda p: p.type.lower() == "personal", profile_list)))
            self.business_profile = cast(BusinessProfile,
                                         next(filter(lambda p: p.type.lower() == "business", profile_list)))
        else:
            self.personal_profile._initialize()
            self.business_profile._initialize()

    @staticmethod
    def get(wise_account_type: WiseAccountType) -> "WiseAccount":
        environmental_variable: EnvironmentSecret = EnvironmentSecret.WISE_PRIMARY_ACCOUNT_API_KEY if wise_account_type == WiseAccountType.PRIMARY else EnvironmentSecret.WISE_SECONDARY_ACCOUNT_API_KEY
        http_session: SyncHTTPSession = SyncHTTPSession(constants.URL, {
            "Authorization": f"Bearer {common.get_environmental_secret(environmental_variable)}"})

        wise_account: WiseAccount = WiseAccount.model_construct(type=wise_account_type, personal_profile=None,
                                                                business_profile=None)
        wise_account._http_session = http_session
        wise_account._initialize()

        WiseAccount.model_validate(wise_account)
        return wise_account


class Profile(DataClass):
    id: int
    type: str
    cash_account_list: List["CashAccount"] | None = None
    reserve_account_list: List["ReserveAccount"] | None = None
    recipient_list: List["Recipient"] | None = None
    debit_card_list: List["DebitCard"] | None = None
    wise_account: WiseAccount = Field(exclude=True)

    @property
    def http_session(self) -> SyncHTTPSession:
        return self.wise_account.http_session

    def _initialize(self) -> None:
        cash_account_to_update_list: List["CashAccount"] = CashAccount.get_all(self)
        reserve_account_to_update_list: List["ReserveAccount"] = ReserveAccount.get_all(self)
        recipient_to_update_list: List["Recipient"] = Recipient.get_all(self)

        if self.cash_account_list is None:
            self.cash_account_list = cash_account_to_update_list
        else:
            for original_cash_account in self.cash_account_list:
                try:
                    new_cash_account: CashAccount = next(
                        filter(lambda c: (c.id == original_cash_account.id), cash_account_to_update_list))
                    original_cash_account.__dict__.update(new_cash_account.model_dump())
                    cash_account_to_update_list.remove(new_cash_account)
                except StopIteration:
                    self.cash_account_list.remove(original_cash_account)

            [self.cash_account_list.append(new_cash_account) for new_cash_account in cash_account_to_update_list]  # type: ignore[func-returns-value]

        if self.reserve_account_list is None:
            self.reserve_account_list = reserve_account_to_update_list
        else:
            for original_reserve_account in self.reserve_account_list:
                try:
                    new_reserve_account: ReserveAccount = next(
                        filter(lambda r: (r.id == original_reserve_account.id), reserve_account_to_update_list))
                    original_reserve_account.__dict__.update(new_reserve_account.model_dump())
                    reserve_account_to_update_list.remove(new_reserve_account)
                except StopIteration:
                    self.reserve_account_list.remove(original_reserve_account)

            [self.reserve_account_list.append(new_reserve_account) for new_reserve_account in reserve_account_to_update_list]  # type: ignore[func-returns-value]

        if self.recipient_list is None:
            self.recipient_list = recipient_to_update_list
        else:
            for original_recipient in self.recipient_list:
                try:
                    new_recipient: Recipient = next(
                        filter(lambda r: (r.id == original_recipient.id), recipient_to_update_list))
                    original_recipient.__dict__.update(new_recipient.model_dump())
                    recipient_to_update_list.remove(new_recipient)
                except StopIteration:
                    self.recipient_list.remove(original_recipient)

            [self.recipient_list.append(new_recipient) for new_recipient in recipient_to_update_list]  # type: ignore[func-returns-value]

    def get_cash_account(self, currency: Currency, is_create_if_unavailable: bool = False) -> "CashAccount":
        try:
            return next(filter(lambda c: c.currency == currency, self.cash_account_list))
        except StopIteration:
            if is_create_if_unavailable:
                return CashAccount.open(self, currency)
            else:
                raise CashAccountNotFoundException(f"Currency not found: \n"
                                                   f"Profile: {self.__class__.__name__}"
                                                   f"Currency: {currency.value}")

    def get_reserve_account(self, account_name: str, currency: Currency,
                            is_create_if_unavailable: bool = False) -> "ReserveAccount":

        try:
            return next(filter(lambda r: r.name == account_name and r.currency == currency, self.reserve_account_list))
        except StopIteration:
            if is_create_if_unavailable:
                return ReserveAccount.open(self, account_name, currency)
            else:
                raise ReserveAccountNotFoundException(f"Currency not found: \n"
                                                      f"Profile: {self.__class__.__name__}"
                                                      f"Reserve Account Name: {account_name}")

    def get_recipient(self, account_number: str) -> "Recipient":

        try:
            return next(filter(lambda r: r.account_number == account_number, self.recipient_list))
        except StopIteration:
            raise RecipientNotFoundException(f"Recipient not found: \n"
                                             f"Profile: {self.__class__.__name__}"
                                             f"Account Number: {account_number}")

    @common.only_in_dev
    def _complete_all_transfers(self) -> None:
        for cash_account in self.cash_account_list:
            http_response: HTTPResponse = self.http_session.get(
                f"{constants.ENDPOINT__TRANSFER__GET_ALL.replace('$profileId', str(self.id))}&status=processing&createdDateStart=2021-01-01&limit=200")
            [cash_account._simulate_completed_transfer(data["id"]) for data in http_response.data]

    @staticmethod
    def get_all(wise_account: WiseAccount) -> List["Profile"]:
        http_response: HTTPResponse = wise_account.http_session.get(constants.ENDPOINT__PROFILE__GET_ALL)
        profile_list: List["Profile"] = []

        for data in http_response.data:
            profile: Profile = Profile(id=data["id"], type=data["type"], wise_account=wise_account)
            profile._initialize()
            profile_list.append(profile)

        return profile_list

    @common.only_in_dev
    async def _reset(self) -> None:
        self._complete_all_transfers()

        for reserve_account in self.reserve_account_list:
            await reserve_account._set_balance(Decimal("0"))
            reserve_account.close()

        [await cash_account._set_balance(Decimal("0")) for cash_account in self.cash_account_list]


class PersonalProfile(Profile):
    pass


class BusinessProfile(Profile):
    pass


class Transaction(DataClass):
    account: "Account" = Field(exclude=True)
    timestamp: datetime.datetime
    type: TransactionType
    description: str
    amount: Decimal
    third_party: Union[str, "ReserveAccount", "Recipient"]


class Account(DataClass):
    id: int
    name: str | None
    currency: Currency
    balance: Decimal
    profile: Profile = Field(exclude=True)

    @property
    def http_session(self) -> SyncHTTPSession:
        return self.profile.http_session

    def close(self) -> None:
        if self.balance != Decimal("0"):
            raise OperationNotSupportedException(f"Cannot close account due to non-zero account balance:\n"
                                                 f"Account Name: {self.name}\n"
                                                 f"Currency: {self.currency.value}\n"
                                                 f"Balance: {'{:,}'.format(self.balance)}")

        self.http_session.delete(
            constants.ENDPOINT__BALANCE__CLOSE.replace("$profileId", str(self.profile.id)).replace("$balanceId",
                                                                                                   str(self.id)))
        self.profile.wise_account._initialize()

    def get_transactions(self, from_time: datetime.datetime | None = None, to_time: datetime.datetime | None = None,
                         number_of_past_hours: int | None = None) -> List["Transaction"]:
        number_of_past_hours = 24 if number_of_past_hours is None else number_of_past_hours
        if from_time is None:
            from_time = datetime.datetime.now() - datetime.timedelta(days=number_of_past_hours)

        if to_time is None:
            to_time = datetime.datetime.now()

        response: HTTPResponse = self.http_session.get(
            constants.ENDPOINT__BALANCE__GET_TRANSACTIONS.replace("$profileId", str(self.profile.id)).replace(
                "$balanceId", str(self.id)), query_params={
                "currency": self.currency.value,
                "intervalStart": f"{from_time.astimezone(datetime.timezone.utc).replace(microsecond=0).isoformat().split('+')[0]}Z",
                "intervalEnd": f"{to_time.astimezone(datetime.timezone.utc).replace(microsecond=0).isoformat().split('+')[0]}Z",
                "type": "COMPACT"
            })

        transaction_list: List[Transaction] = []
        for data in response.data["transactions"]:
            transaction_type: TransactionType = TransactionType(data["details"]["type"])
            third_party: str | ReserveAccount | Recipient

            if transaction_type == TransactionType.DEPOSIT:
                third_party = data["details"]["senderName"] + " | " + data["details"]["senderAccount"]

            elif transaction_type == TransactionType.TRANSFER:
                try:
                    third_party = self.profile.get_recipient(data["details"]["recipient"]["bankAccount"])
                except RecipientNotFoundException:
                    third_party = data["details"]["recipient"]["name"] + " | " + data["details"]["recipient"][
                        "bankAccount"]

            elif transaction_type == TransactionType.CARD:
                third_party = data["details"]["merchant"]["name"] + " | " + data["details"]["merchant"]["country"]

            elif transaction_type == TransactionType.CONVERSION:
                description: str = data["details"]["description"]
                reserve_account_name: str = description.split(" to ")[1] if " to " in description else \
                description.split(" from ")[1]
                try:
                    third_party = self.profile.get_reserve_account(reserve_account_name, self.currency)
                except ReserveAccountNotFoundException:
                    third_party = reserve_account_name

            transaction_list.append(Transaction(
                account=self,
                timestamp=data["date"],
                type=transaction_type,
                description=data["details"]["description"],
                amount=Decimal(str(data["amount"]["value"])),
                third_party=third_party
            ))

        return transaction_list

    @staticmethod
    def abstract_open(profile: Profile, account_name: str | None, currency: Currency,
                      is_reserve_account: bool) -> "Account":
        data = {
            "currency": currency.value,
            "type": "SAVINGS" if is_reserve_account else "STANDARD"
        }

        if is_reserve_account:
            data["name"] = account_name

        response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__BALANCE__OPEN.replace("$profileId", str(profile.id)), data=data,
            headers={"X-idempotence-uuid": str(uuid.uuid4())})
        return Account(
            id=response.data["id"],
            name=account_name,
            currency=currency,
            balance=Decimal("0"),
            profile=profile
        )


class CashAccount(Account):

    @common.only_in_dev
    def _simulate_completed_transfer(self, transfer_id: int) -> None:
        url: str = constants.ENDPOINT__SIMULATION__COMPLETE_TRANSFER.replace("$transferId", str(transfer_id))
        try:
            self.http_session.get(url.replace("$status", "processing"))
            self.http_session.get(url.replace("$status", "funds_converted"))
            self.http_session.get(url.replace("$status", "outgoing_payment_sent"))
        except HTTPException:
            pass

    async def transfer(self, to_account: Union["CashAccount", "ReserveAccount", "Recipient"], amount: Decimal,
                       reference: str | None = None, is_amount_in_from_currency: bool = False) -> "Transfer":
        amount = amount.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)

        if isinstance(to_account, ReserveAccount) and self.currency != to_account.currency:
            raise OperationNotSupportedException(
                "Direct inter-currency transfers from a cash account to a reserve account is not supported")

        transfer: Transfer = Transfer.model_construct()
        if isinstance(to_account, CashAccount):
            transfer = Transfer.intra_cash_account_transfer(self.profile, self, to_account, amount,
                                                            is_amount_in_from_currency)
            await DiscordDefaults.send_message(AortaTextChannels.WISE.value, f"**Intra-Account Transfer**:\n"
                                                                             f"Timestamp: {get_timestamp_string(datetime.datetime.now())}\n"
                                                                             f"From: *{self.currency.value}*\n"
                                                                             f"To: *{to_account.currency.value}*\n"
                                                                             f"Amount: *{self.currency.value} {'{:,}'.format(amount)}*\n"
                                               )

        elif isinstance(to_account, ReserveAccount):
            transfer = Transfer.cash_to_savings_account_transfer(self.profile, self, to_account, amount)
            await DiscordDefaults.send_message(AortaTextChannels.WISE.value, f"**Intra-Account Transfer**:\n"
                                                                             f"Timestamp: {get_timestamp_string(datetime.datetime.now())}\n"
                                                                             f"From: *{self.currency.value}*\n"
                                                                             f"To: *{to_account.name}*\n"
                                                                             f"Amount: *{self.currency.value} {'{:,}'.format(amount)}*\n")

        elif isinstance(to_account, Recipient):
            transfer = Transfer.cash_to_third_party_cash_account_transfer(self.profile, self, to_account, amount,
                                                                          "" if reference is None else reference,
                                                                          is_amount_in_from_currency)
            await DiscordDefaults.send_message(AortaTextChannels.WISE.value, f"**Third-Party Transfer**:\n"
                                                                             f"Timestamp: {get_timestamp_string(datetime.datetime.now())}\n"
                                                                             f"From: *{self.currency.value}*\n"
                                                                             f"To: *{to_account.account_holder_name}*\n"
                                                                             f"Amount: *{self.currency.value} {'{:,}'.format(amount)}*\n")

            if not common.is_production_environment():
                self._simulate_completed_transfer(transfer.id)

        self.profile.wise_account._initialize()
        Transfer.model_validate(transfer)
        return transfer

    @common.only_in_dev
    def _simulate_top_up(self, amount: Decimal) -> None:
        amount = amount.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)
        if not common.is_development_environment():
            raise OperationNotSupportedException("Simulations can only be done in a development environment")

        self.profile.http_session.post(constants.ENDPOINT__SIMULATION__TOP_UP, {
            "profileId": self.profile.id,
            "balanceId": self.id,
            "currency": self.currency.value,
            "amount": float(amount)
        })
        self.profile.wise_account._initialize()

    @common.only_in_dev
    async def _set_balance(self, amount: Decimal) -> None:
        if self.balance > amount:
            await self._set_maximum_balance(amount)
        else:
            self._set_minimum_balance(amount)

    @common.only_in_dev
    def _set_minimum_balance(self, amount: Decimal) -> None:
        if self.balance < amount:
            self._simulate_top_up(amount - self.balance)

    @common.only_in_dev
    async def _set_maximum_balance(self, amount: Decimal) -> None:
        if self.balance <= amount:
            return

        maximum_transfer_amount: Decimal = Decimal(1_000_000)
        amount_to_deduct: Decimal = min(self.balance - amount, maximum_transfer_amount)
        recipient: Recipient = self.profile.get_recipient("1234567890")

        if amount_to_deduct < Decimal("10"):
            amount_to_deduct = amount_to_deduct + Decimal("10")
            self._set_minimum_balance(self.balance + Decimal("10"))

        await self.transfer(recipient, amount_to_deduct, is_amount_in_from_currency=True)

        if self.balance > amount:
            await self._set_maximum_balance(amount)

    @staticmethod
    def get_all(profile: Profile) -> List["CashAccount"]:
        response: HTTPResponse = profile.http_session.get(
            constants.ENDPOINT__ACCOUNT__GET_ALL__CASH_ACCOUNT.replace("$profileId", str(profile.id)))
        return [CashAccount(
            id=data["id"],
            name=data["name"],
            currency=Currency(data["cashAmount"]["currency"]),
            balance=Decimal(str(data["cashAmount"]["value"])),
            profile=profile
        ) for data in response.data]

    @staticmethod
    def open(profile: Profile, currency: Currency) -> "CashAccount":
        account: Account = Account.abstract_open(profile, None, currency, False)
        cash_account: CashAccount = CashAccount.model_construct(**account.model_dump())
        cash_account.profile = profile

        profile.cash_account_list.append(cash_account)
        return cash_account


class ReserveAccount(Account):

    async def transfer(self, to_account: "CashAccount", amount: Decimal) -> "Transfer":
        amount = amount.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)

        if self.currency != to_account.currency:
            raise OperationNotSupportedException(
                "Direct inter-currency transfers from a reserve account is not supported")

        transfer: Transfer = Transfer.savings_to_cash_account_transfer(self.profile, self, to_account, amount)
        await DiscordDefaults.send_message(AortaTextChannels.WISE.value, f"**Intra-Account Transfer**:\n"
                                                                         f"*Timestamp*: {get_timestamp_string(datetime.datetime.now())}\n"
                                                                         f"*From*: {self.name}\n"
                                                                         f"*To*: {to_account.currency.value}\n"
                                                                         f"*Amount*: {self.currency.value} {'{:,}'.format(amount)}\n")

        self.profile.wise_account._initialize()
        return transfer

    @common.only_in_dev
    async def _simulate_top_up(self, amount: Decimal) -> None:
        amount = amount.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP)
        cash_account: CashAccount = self.profile.get_cash_account(self.currency, True)
        cash_account._simulate_top_up(amount)
        await cash_account.transfer(self, amount)
        self.profile.wise_account._initialize()
        self.balance = (
            self.profile.wise_account.personal_profile.get_reserve_account(self.name, self.currency)).balance

    @common.only_in_dev
    async def _set_balance(self, amount: Decimal) -> None:
        if self.balance == amount:
            return
        elif self.balance < amount:
            await self._set_minimum_balance(amount)
        else:
            await self._set_maximum_balance(amount)

    @common.only_in_dev
    async def _set_minimum_balance(self, amount: Decimal) -> None:
        if self.balance >= amount:
            return

        amount_to_top_up: Decimal = amount - self.balance
        cash_account: CashAccount = self.profile.get_cash_account(self.currency, True)
        cash_account._simulate_top_up(amount_to_top_up)
        await cash_account.transfer(self, amount_to_top_up)

    @common.only_in_dev
    async def _set_maximum_balance(self, amount: Decimal) -> None:
        if self.balance <= amount:
            return

        maximum_transfer_amount: Decimal = Decimal(7_000_000)
        amount_to_deduct: Decimal = self.balance - amount
        cash_account: CashAccount = self.profile.get_cash_account(self.currency, True)
        hkd_account: CashAccount = self.profile.get_cash_account(common.Currency.HKD, True)
        await self.transfer(cash_account, amount_to_deduct)

        if amount_to_deduct > maximum_transfer_amount:
            await cash_account.transfer(hkd_account, maximum_transfer_amount)
            self._set_maximum_balance(amount)
        else:
            quote: Quote = Quote.get_quote(self.profile, cash_account, hkd_account, amount_to_deduct, True)
            await cash_account.transfer(hkd_account, quote.from_amount, is_amount_in_from_currency=True)

    @staticmethod
    def get_all(profile: Profile) -> List["ReserveAccount"]:
        response: HTTPResponse = profile.http_session.get(
            constants.ENDPOINT__ACCOUNT__GET_ALL__RESERVE_ACCOUNT.replace("$profileId", str(profile.id)))
        return [ReserveAccount(
            id=data["id"],
            name=data["name"],
            currency=Currency(data["cashAmount"]["currency"]),
            balance=Decimal(str(data["cashAmount"]["value"])),
            profile=profile,
        ) for data in response.data]

    @staticmethod
    def open(profile: Profile, account_name: str, currency: Currency) -> "ReserveAccount":
        account: Account = Account.abstract_open(profile, account_name, currency, True)
        reserve_account: ReserveAccount = ReserveAccount.model_construct(**account.model_dump())
        reserve_account.profile = profile

        profile.reserve_account_list.append(reserve_account)
        return reserve_account


class Recipient(DataClass):
    id: int
    account_holder_name: str
    currency: Currency
    is_self_owned: bool
    account_number: str
    _http_session: SyncHTTPSession = PrivateAttr()

    @staticmethod
    def get_all(profile: Profile) -> List["Recipient"]:
        response: HTTPResponse = profile.http_session.get(
            constants.ENDPOINT__RECIPIENT__GET_ALL.replace("$profileId", str(profile.id)))
        raw_recipient_list: List[Dict[str, Any]] = list(
            filter(lambda d: d["details"]["accountNumber"] is not None or d["details"]["iban"] is not None,
                   response.data))

        return [Recipient(
            id=data["id"],
            account_holder_name=data["accountHolderName"],
            currency=Currency(data["currency"]),
            is_self_owned=data["ownedByCustomer"],
            account_number=data["details"]["accountNumber"] if data["details"]["accountNumber"] is not None else
            data["details"]["iban"],
        ) for data in raw_recipient_list]


class Quote(DataClass):
    id: str
    from_currency: Currency
    to_currency: Currency
    from_amount: Decimal
    to_amount: Decimal
    exchange_rate: Decimal
    profile: Profile

    @staticmethod
    def get_quote(profile: Profile, from_account: CashAccount | ReserveAccount,
                  to_account: CashAccount | ReserveAccount | Recipient, amount: Decimal,
                  is_amount_in_from_currency: bool = False) -> "Quote":
        response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__QUOTE__GET.replace("$profileId", str(profile.id)), data={
                "sourceCurrency": from_account.currency.value,
                "targetCurrency": to_account.currency.value,
                f"{'sourceAmount' if is_amount_in_from_currency else 'targetAmount'}": float(amount),
                "payOut": "BALANCE",
            })

        payment_option: Dict[str, Any] = next(
            filter(lambda p: p["payIn"] == "BALANCE", response.data["paymentOptions"]))
        return Quote(
            id=response.data["id"],
            from_currency=Currency(payment_option["sourceCurrency"]),
            to_currency=Currency(str(payment_option["targetCurrency"])),
            from_amount=Decimal(str(payment_option["sourceAmount"])),
            to_amount=Decimal(str(payment_option["targetAmount"])),
            exchange_rate=Decimal(str(response.data["rate"])),
            profile=profile
        )


class TransferType(Enum):
    CASH_TO_SAVINGS: int = auto()
    SAVINGS_TO_CASH: int = auto()
    CASH_TO_THIRD_PARTY: int = auto()
    SAVINGS_TO_THIRD_PARTY: int = auto()
    INTRA_CASH: int = auto()
    INTRA_SAVINGS: int = auto()


class Transfer(DataClass):
    id: int
    from_account: CashAccount | ReserveAccount
    to_account: CashAccount | ReserveAccount | Recipient
    from_amount: Decimal
    to_amount: Decimal
    reference: str | None
    transfer_type: TransferType | None

    @staticmethod
    def intra_cash_account_transfer(profile: Profile, from_account: CashAccount, to_account: CashAccount,
                                    amount: Decimal, is_amount_in_from_currency: bool = False) -> "Transfer":
        quote: Quote = Quote.get_quote(profile, from_account, to_account, amount, is_amount_in_from_currency)
        response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__BALANCE__MOVE_MONEY_BETWEEN_BALANCES.replace("$profileId", str(profile.id)),
            data={"quoteId": quote.id},
            headers={"X-idempotence-uuid": str(uuid.uuid4())})

        return Transfer(
            id=response.data["id"],
            from_account=from_account,
            from_amount=Decimal(str(response.data["sourceAmount"]["value"])),
            to_account=to_account,
            to_amount=Decimal(str(response.data["targetAmount"]["value"])),
            reference=None,
            transfer_type=TransferType.INTRA_CASH,
        )

    @staticmethod
    def cash_to_savings_account_transfer(profile: Profile, from_account: CashAccount, to_account: ReserveAccount,
                                         amount: Decimal) -> "Transfer":
        data = {
            "sourceBalanceId": from_account.id,
            "targetBalanceId": to_account.id
        }

        if from_account.currency != to_account.currency:
            quote: Quote = Quote.get_quote(profile, from_account, to_account, amount)
            data["quoteId"] = cast(int, quote.id)
        else:
            data["amount"] = {  # type: ignore[assignment]
                "value": float(amount),
                "currency": to_account.currency.value
            }

        response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__BALANCE__MOVE_MONEY_BETWEEN_BALANCES.replace("$profileId", str(profile.id)), data=data,
            headers={"X-idempotence-uuid": str(uuid.uuid4())})

        return Transfer(
            id=response.data["id"],
            from_account=from_account,
            from_amount=Decimal(str(response.data["sourceAmount"]["value"])),
            to_account=to_account,
            to_amount=Decimal(str(response.data["targetAmount"]["value"])),
            reference=None,
            transfer_type=TransferType.CASH_TO_SAVINGS,
        )

    @staticmethod
    def cash_to_third_party_cash_account_transfer(profile: Profile, from_account: CashAccount, to_account: Recipient,
                                                  amount: Decimal, reference: str | None = None,
                                                  is_amount_in_from_currency: bool = False) -> "Transfer":
        quote: Quote = Quote.get_quote(profile, from_account, to_account, amount, is_amount_in_from_currency)
        data: Dict[str, Any] = {
            "targetAccount": to_account.id,
            "quoteUuid": quote.id,
            "customerTransactionId": str(uuid.uuid4()),
            "details": {
                "reference": "" if reference is None else reference,
            }
        }

        create_transfer_response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__TRANSFER__CREATE_THIRD_PARTY_TRANSFER, data=data)
        profile.http_session.post(
            constants.ENDPOINT__TRANSFER__FUND_THIRD_PARTY_TRANSFER.replace("$profileId", str(profile.id)).replace(
                "$transferId", str(create_transfer_response.data["id"])),
            data={"type": "BALANCE"})

        return Transfer(
            id=create_transfer_response.data["id"],
            from_account=from_account,
            from_amount=Decimal(str(create_transfer_response.data["sourceValue"])),
            to_account=to_account,
            to_amount=Decimal(str(create_transfer_response.data["targetValue"])),
            reference=None,
            transfer_type=TransferType.CASH_TO_THIRD_PARTY,
        )

    @staticmethod
    def savings_to_cash_account_transfer(profile: Profile, from_account: ReserveAccount, to_account: CashAccount,
                                         amount: Decimal) -> "Transfer":
        data = {
            "amount": {
                "value": float(amount),
                "currency": from_account.currency.value
            },
            "sourceBalanceId": from_account.id,
            "targetBalanceId": to_account.id,
        }

        response: HTTPResponse = profile.http_session.post(
            constants.ENDPOINT__BALANCE__MOVE_MONEY_BETWEEN_BALANCES.replace("$profileId", str(profile.id)), data=data,
            headers={"X-idempotence-uuid": str(uuid.uuid4())})

        return Transfer(
            id=response.data["id"],
            from_account=from_account,
            from_amount=Decimal(str(response.data["sourceAmount"]["value"])),
            to_account=to_account,
            to_amount=Decimal(str(response.data["targetAmount"]["value"])),
            reference=None,
            transfer_type=TransferType.SAVINGS_TO_CASH,
        )


class DebitCard(DataClass):
    profile: Profile
    token: str
    expiry_date: datetime.datetime
    bank_identification_number: str

    # TODO: Find out why this endpoint returns a 403 (Unauthorized)
    @staticmethod
    def get_all(profile: Profile) -> List["DebitCard"]:
        response: HTTPResponse = profile.http_session.get(
            constants.ENDPOINT__DEBIT_CARD__GET_ALL.replace("$profileId", str(profile.id)))
        return [DebitCard(
            profile=profile,
            token=data["token"],
            expiry_date=datetime.datetime.fromisoformat(data["expiryDate"]),
            bank_identification_number=data["bankIdentificationNumber"]
        ) for data in response.data["cards"]]


class AccountDebit(DataClass):
    wise_id: int
    is_attempted: bool
    is_successful: bool
    timestamp: datetime.datetime

    @staticmethod
    def get_from_request_data(request_data: Dict[str, Any]) -> "AccountDebit":
        return AccountDebit(wise_id=request_data["data"]["resource"]["id"],
                            is_attempted=request_data["data"]["current_state"] == "incoming_payment_waiting",
                            is_successful=request_data["data"]["current_state"] == "outgoing_payment_sent",
                            timestamp=request_data["data"]["occurred_at"])


class AccountCredit(DataClass):
    wise_id: int
    account: CashAccount
    transaction: Transaction | None
    account_balance: Decimal
    timestamp: datetime.datetime

    @staticmethod
    def get_from_request_data(request_data: Dict[str, Any], wise_account: WiseAccount) -> "AccountCredit":
        personal_profile: PersonalProfile = wise_account.personal_profile
        cash_account: CashAccount = personal_profile.get_cash_account(Currency(request_data["data"]["currency"]))
        timestamp: datetime.datetime = common.get_timestamp_from_string(request_data["data"]["occurred_at"], "UTC")

        try:
            transaction: Transaction | None = next(
                filter(lambda t: int(t.timestamp.timestamp()) == int(timestamp.timestamp()),
                       cash_account.get_transactions()))
        except StopIteration:
            transaction = None

        return AccountCredit(wise_id=request_data["data"]["resource"]["id"],
                             account=cash_account,
                             transaction=transaction,
                             account_balance=Decimal(request_data["data"]["post_transaction_balance_amount"]),
                             timestamp=request_data["data"]["occurred_at"])


class WiseWebhook:

    @classmethod
    async def get_balance_update_object(cls, request_data: Dict[str, Any],
                                        wise_account: WiseAccount) -> AccountDebit | AccountCredit | None:
        if request_data["event_type"] == "transfers#state-change":
            return AccountDebit.get_from_request_data(request_data)
        elif request_data["event_type"] == "balances#credit":
            return AccountCredit.get_from_request_data(request_data, wise_account)

        return None


WiseAccount.model_rebuild()
Profile.model_rebuild()
PersonalProfile.model_rebuild()
BusinessProfile.model_rebuild()
Account.model_rebuild()
DebitCard.model_rebuild()
Transaction.model_rebuild()
AccountDebit.model_rebuild()
AccountCredit.model_rebuild()



import json
from abc import abstractmethod
from dataclasses import dataclass
from typing import Dict, Any, List, cast

import httpx
from httpx import Response, Cookies, Headers, URL, AsyncClient, Client
from pydantic import BaseModel

from sirius import application_performance_monitoring, common
from sirius.application_performance_monitoring import Operation
from sirius.common import DataClass
from sirius.http_requests.exceptions import ClientSideException, ServerSideException


@dataclass
class HTTPResponse:
    response: Response
    response_code: int
    is_successful: bool
    headers: Headers
    data: Dict[Any, Any] | None = None
    response_text: str | None = None
    cookies: Cookies | None = None

    def __init__(self, response: Response, *args: List[Any], **kwargs: Dict[str, Any]) -> None:
        self.response = response
        self.response_code = self.response.status_code
        self.is_successful = 200 <= self.response_code < 300
        self.response_text = self.response.text
        self.headers = response.headers
        self.cookies = response.cookies

        if self.response_text is not None and self.response_text != "":
            self.data = self.response.json()

        super().__init__(*args, **kwargs)


class HTTPSession:
    host: str

    @staticmethod
    def raise_http_exception(http_response: HTTPResponse) -> None:
        error_message: str = f"HTTP Exception\n" \
                             f"URL: {str(http_response.response.url)}\n" \
                             f"Headers: {str(http_response.headers)}\n" \
                             f"Method: {http_response.response.request.method.upper()}\n" \
                             f"Response Code: {http_response.response_code}\n" \
                             f"Response Text: {http_response.response_text}"

        if 400 <= http_response.response_code < 500:
            raise ClientSideException(error_message, data={"http_response": http_response})
        else:
            raise ServerSideException(error_message, data={"http_response": http_response})


class AsyncHTTPSession(HTTPSession):
    client: AsyncClient
    _instance_list: List["AsyncHTTPSession"] = []

    def __new__(cls, url_str: str, headers: Dict[str, Any] | None = None) -> "AsyncHTTPSession":
        host: str = URL(url_str).host
        instance: AsyncHTTPSession | None = None

        for i in cls._instance_list:
            if i.host == host and (headers is None or common.is_dict_include_another_dict(cast(Dict[str, Any], headers), dict(i.client.headers))):
                instance = i

        if instance is None:
            instance = super().__new__(cls)
            instance.host = host
            instance.client = httpx.AsyncClient()

            if headers is not None:
                instance.client.headers.update(headers)

            cls._instance_list.append(instance)

        if instance.client.is_closed:
            instance.client = httpx.AsyncClient()

        return instance

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "GET")
    async def get(self, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(await self.client.get(url, params=query_params, headers=headers))
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "PUT")
    async def put(self, url: str, data: Dict[str, Any], headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(await self.client.put(url, data=data, headers=headers))
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "POST")
    async def post(self, url: str, data: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None, is_form_url_encoded: bool = False) -> HTTPResponse:
        data_string: str | None = None
        if data is not None:
            data_string = json.dumps(data)

            if headers is None:
                headers = {}

            if "content-type" not in headers and not is_form_url_encoded:
                headers["content-type"] = "application/json"

        http_response: HTTPResponse = HTTPResponse(await self.client.post(url, data=data, headers=headers)) if is_form_url_encoded else HTTPResponse(await self.client.post(url, data=data_string, headers=headers))  # type: ignore[arg-type]
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response

    # @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "DELETE")
    async def delete(self, url: str, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(await self.client.delete(url, headers=headers))
        if not http_response.is_successful:
            AsyncHTTPSession.raise_http_exception(http_response)

        return http_response


class SyncHTTPSession(HTTPSession):
    _instance_list: List["SyncHTTPSession"] = []
    client: Client
    host: str

    def __new__(cls, url_str: str, headers: Dict[str, Any] | None = None) -> "SyncHTTPSession":
        host: str = URL(url_str).host
        instance: SyncHTTPSession | None = None

        for i in cls._instance_list:
            if i.host == host and (headers is None or common.is_dict_include_another_dict(cast(Dict[str, Any], headers), dict(i.client.headers))):
                instance = i

        if instance is None:
            instance = super().__new__(cls)
            instance.host = host
            instance.client = httpx.Client()

            if headers is not None:
                instance.client.headers.update(headers)

            cls._instance_list.append(instance)

        if instance.client.is_closed:
            instance.client = httpx.Client()

        return instance

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "GET")
    def get(self, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None, ) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(self.client.get(url, params=query_params, headers=headers, timeout=60))
        if not http_response.is_successful:
            SyncHTTPSession.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "PUT")
    def put(self, url: str, data: Dict[str, Any], headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(self.client.put(url, data=data, headers=headers, timeout=60))
        if not http_response.is_successful:
            SyncHTTPSession.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "POST")
    def post(self, url: str, data: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        data_string: str | None = None
        if data is not None:
            data_string = json.dumps(data)

            if headers is None:
                headers = {}
            headers["content-type"] = "application/json"

        http_response: HTTPResponse = HTTPResponse(self.client.post(url, data=data_string, headers=headers, timeout=60))  # type: ignore[arg-type]
        if not http_response.is_successful:
            SyncHTTPSession.raise_http_exception(http_response)

        return http_response

    @application_performance_monitoring.transaction(Operation.HTTP_REQUEST, "DELETE")
    def delete(self, url: str, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        http_response: HTTPResponse = HTTPResponse(self.client.delete(url, headers=headers, timeout=60))
        if not http_response.is_successful:
            SyncHTTPSession.raise_http_exception(http_response)

        return http_response


class HTTPModel(DataClass):

    @abstractmethod
    def __init__(self, **data: Any):
        super().__init__(**data)

    @staticmethod
    async def get_one(cls: type, url: str, query_params: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> DataClass:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.get(url=url, query_params=query_params)
        return cls(**response.data)  # type: ignore[return-value]

    @staticmethod
    async def get_multiple(cls: type, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> List[DataClass]:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.get(url=url, query_params=query_params, headers=headers)
        return [cls(**data) for data in response.data]  # type: ignore[misc]

    @staticmethod
    async def post_return_one(cls: type, url: str, data: Dict[Any, Any] | None = None, headers: Dict[str, Any] | None = None, http_session: AsyncHTTPSession | None = None) -> DataClass:
        if http_session is None:
            http_session = AsyncHTTPSession(url)

        if not issubclass(cls, BaseModel):
            raise ServerSideException(f"{cls.__name__} is not a Pydantic subclass")

        response: HTTPResponse = await http_session.post(url=url, data=data, headers=headers)
        return cls(**response.data)  # type: ignore[return-value]



from sirius.exceptions import SiriusException, SDKClientException


class WiseException(SiriusException):
    pass


class CashAccountNotFoundException(WiseException, SDKClientException):
    pass


class ReserveAccountNotFoundException(WiseException, SDKClientException):
    pass


class RecipientNotFoundException(WiseException, SDKClientException):
    pass



import base64
import hashlib
import json
from typing import Dict, Any

from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.keyvault.keys import KeyClient, KeyVaultKey, KeyType
from azure.keyvault.keys.crypto import CryptographyClient, SignatureAlgorithm, SignResult, EncryptionAlgorithm
from azure.keyvault.secrets import SecretClient

from sirius import common
from sirius.constants import EnvironmentVariable
from sirius.exceptions import SDKClientException

secret_cache: Dict[str, str] = {}
key_cache: Dict[str, KeyVaultKey] = {}


class AzureKeyVault:
    secret_client: SecretClient | None = None
    key_client: KeyClient | None = None

    @classmethod
    def authenticate(cls) -> None:
        if cls.secret_client is not None and cls.key_client is not None:
            return

        cls.secret_client = SecretClient(vault_url=common.get_environmental_variable(EnvironmentVariable.AZURE_KEY_VAULT_URL), credential=DefaultAzureCredential())
        cls.key_client = KeyClient(vault_url=common.get_environmental_variable(EnvironmentVariable.AZURE_KEY_VAULT_URL), credential=DefaultAzureCredential())

    @classmethod
    def get(cls, key: str) -> str:
        global secret_cache
        if key in secret_cache:
            return secret_cache[key]

        cls.authenticate()
        try:
            value: str = cls.secret_client.get_secret(key).value
        except ResourceNotFoundError:
            raise SDKClientException("Secret not found: " + key)

        secret_cache[key] = value
        return value

    @classmethod
    def set(cls, key: str, value: str) -> None:
        cls.authenticate()
        cls.secret_client.set_secret(key, value)

        global secret_cache
        secret_cache[key] = value

    @classmethod
    def delete(cls, key: str) -> None:
        cls.authenticate()
        cls.secret_client.begin_delete_secret(key).result()

        global cacheecr
        if key in secret_cache:
            secret_cache.pop(key)

    @classmethod
    def get_key(cls, key_name: str, is_generate_if_unavailable: bool = False) -> KeyVaultKey:
        global key_cache
        if key_name in key_cache:
            return key_cache[key_name]

        cls.authenticate()

        try:
            azure_key: KeyVaultKey = cls.key_client.get_key(key_name)
        except ResourceNotFoundError:
            if is_generate_if_unavailable:
                return AzureKeyVault.create_key(key_name)
            else:
                raise SDKClientException("Key not found: " + key_name)

        key_cache[key_name] = azure_key
        return azure_key

    @classmethod
    def create_key(cls, key_name: str) -> KeyVaultKey:
        global key_cache
        cls.authenticate()

        azure_key: KeyVaultKey = cls.key_client.create_key(key_name, size=4096, key_type=KeyType.rsa)
        key_cache[key_name] = azure_key
        return azure_key

    @classmethod
    def delete_key(cls, key_name: str) -> None:
        global key_cache
        cls.authenticate()
        cls.key_client.begin_delete_key(key_name).result()
        key_cache.pop(key_name)


class AzureKeyVaultCryptography:
    @staticmethod
    def get_encrypted_data(key_name: str, data: bytes) -> bytes:
        azure_key: KeyVaultKey = AzureKeyVault.get_key(key_name, True)
        crypto_client: CryptographyClient = CryptographyClient(azure_key, DefaultAzureCredential())
        return crypto_client.encrypt(EncryptionAlgorithm.rsa_oaep_256, data).ciphertext

    @staticmethod
    def get_decrypted_data(key_name: str, encrypted_data: bytes) -> bytes:
        azure_key: KeyVaultKey = AzureKeyVault.get_key(key_name, True)
        crypto_client: CryptographyClient = CryptographyClient(azure_key, DefaultAzureCredential())
        return crypto_client.decrypt(EncryptionAlgorithm.rsa_oaep_256, encrypted_data).plaintext

    @staticmethod
    def get_jwt_token(key_name: str, payload: Dict[str, Any], header: Dict[str, Any] | None = None) -> str:
        header = {} if header is None else header
        header["alg"] = "RS256"
        header["typ"] = "JWT"

        azure_key: KeyVaultKey = AzureKeyVault.get_key(key_name, True)
        crypto_client: CryptographyClient = CryptographyClient(azure_key, DefaultAzureCredential())
        header_encoded: str = base64.urlsafe_b64encode(json.dumps(header).encode()).decode().rstrip("=")
        payload_encoded: str = base64.urlsafe_b64encode(json.dumps(payload).encode()).decode().rstrip("=")

        hash_var = hashlib.sha256()
        hash_var.update(f"{header_encoded}.{payload_encoded}".encode("utf-8"))

        sign_result: SignResult = crypto_client.sign(SignatureAlgorithm.rs256, hash_var.digest())
        signature: str = base64.urlsafe_b64encode(sign_result.signature).decode().rstrip("=")
        return f"{header_encoded}.{payload_encoded}.{signature}"



ACQUIRE_ACCESS_TOKEN__POLLING_SLEEP_SECONDS: int = 1
ACQUIRE_ACCESS_TOKEN__POLLING_TIMEOUT_SECONDS: int = 5 * 60
AUTHENTICATION_KEY_NAME: str = "KEY-AUTHENTICATION"
DEFAULT_ACCESS_TOKEN_VALIDITY_SECONDS: int = 3600



import asyncio
from typing import Callable, Dict, Any

import pytz
from apscheduler.schedulers.asyncio import AsyncIOScheduler


class AsynchronousScheduler:
    _instance: AsyncIOScheduler | None = None

    @classmethod
    async def add_job(cls, func: Callable, **kwargs: Dict[str, Any]) -> None:
        """"Parameter reference: https://apscheduler.readthedocs.io/en/stable/modules/triggers/cron.html#module-apscheduler.triggers.cron"""
        cls._instance = AsyncIOScheduler(timezone=pytz.timezone("Pacific/Auckland")) if cls._instance is None else cls._instance
        cls._instance.add_job(func, "cron", **kwargs)
        if not cls._instance.running:
            cls._instance.start()



import datetime
from typing import Union, cast, List, Dict, Any

import motor
from bson import ObjectId
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorDatabase, AsyncIOMotorCollection

from sirius import common
from sirius.common import DataClass
from sirius.constants import EnvironmentSecret

client: AsyncIOMotorClient | None = None  # type: ignore[valid-type]
db: AsyncIOMotorDatabase | None = None  # type: ignore[valid-type]


async def initialize() -> None:
    global client, db
    client = motor.motor_asyncio.AsyncIOMotorClient(
        f"{common.get_environmental_secret(EnvironmentSecret.MONGO_DB_CONNECTION_STRING)}&retryWrites=false",
                                                    uuidRepresentation="standard") if client is None else client
    db = client[common.get_environmental_secret(EnvironmentSecret.APPLICATION_NAME)] if db is None else db


async def drop_collection(collection_name: str) -> None:
    await initialize()
    await cast(AsyncIOMotorDatabase, db).drop_collection(collection_name)  # type: ignore[attr-defined,valid-type]


class DatabaseDocument(DataClass):
    id: ObjectId | None = None
    updated_timestamp: datetime.datetime | None = None
    created_timestamp: datetime.datetime | None = None

    @classmethod
    async def _get_collection(cls) -> AsyncIOMotorCollection:  # type: ignore[valid-type]
        await initialize()
        global db
        return db[cls.__name__]  # type: ignore[index]

    async def save(self) -> None:
        collection: AsyncIOMotorCollection = await self._get_collection()  # type: ignore[valid-type]

        if self.id is None:
            self.created_timestamp = datetime.datetime.now()
            object_id: ObjectId = (await collection.insert_one(self.model_dump(exclude={"id"}))).inserted_id  # type: ignore[attr-defined]
            self.__dict__.update(self.model_dump(exclude={"id"}))
            self.id = object_id
        else:
            self.updated_timestamp = datetime.datetime.now()
            await collection.replace_one({"_id": self.id}, self.model_dump(exclude={"id"}))  # type: ignore[attr-defined]

    async def delete(self) -> None:
        await initialize()
        collection: AsyncIOMotorCollection = await self._get_collection()  # type: ignore[valid-type]
        await collection.delete_one({'_id': self.id})  # type: ignore[attr-defined]

    @classmethod
    def get_model_by_raw_data(cls, raw_data: Dict[Any, Any]) -> "DatabaseDocument":
        object_id = raw_data.pop("_id")
        queried_object: DatabaseDocument = cls(**raw_data)
        queried_object.id = object_id
        return queried_object

    @classmethod
    async def find_by_id(cls, object_id: ObjectId) -> Union["DatabaseDocument", None]:
        await initialize()
        collection: AsyncIOMotorCollection = await cls._get_collection()  # type: ignore[valid-type]
        object_model: Dict[str, Any] = await collection.find_one({'_id': object_id})  # type: ignore[attr-defined]
        return None if object_model is None else cls.get_model_by_raw_data(object_model)

    @classmethod
    async def find_by_query(cls, database_document: "DatabaseDocument", query_limit: int = 100) -> List["DatabaseDocument"]:
        collection: AsyncIOMotorCollection = await cls._get_collection()  # type: ignore[valid-type]
        cursor = collection.find(database_document.model_dump(exclude={"id"}, exclude_none=True))  # type: ignore[attr-defined]
        return [cls.get_model_by_raw_data(document) for document in await cursor.to_list(length=query_limit)]



from sirius.exceptions import ApplicationException


class IAMException(ApplicationException):
    pass


class InvalidAccessTokenException(IAMException):
    pass


class AccessTokenRetrievalTimeoutException(IAMException):
    pass


class AccessTokenRetrievalException(IAMException):
    pass



import asyncio
import base64
import datetime
import hashlib
import time
from typing import Any, Dict, Callable, Union
from urllib.parse import urlencode

import jwt

from sirius import common
from sirius.common import DataClass
from sirius.communication.discord import AortaTextChannels, DiscordDefaults
from sirius.constants import EnvironmentSecret
from sirius.http_requests import AsyncHTTPSession, HTTPResponse, ClientSideException
from sirius.iam import constants
from sirius.iam.exceptions import InvalidAccessTokenException, AccessTokenRetrievalTimeoutException


class AuthenticationFlow(DataClass):
    user_code: str
    device_code: str
    verification_uri: str
    message: str
    expiry_timestamp: datetime.datetime


class MicrosoftEntraIDAuthenticationIDStore:
    store: Dict[str, Any] = {}

    @classmethod
    def add(cls, authentication_id: str, authentication_code: str) -> None:
        cls.store[authentication_id] = authentication_code

    @classmethod
    async def _get_or_wait(cls, authentication_id: str) -> str:
        time_out_timestamp: int = int(time.time()) + constants.ACQUIRE_ACCESS_TOKEN__POLLING_TIMEOUT_SECONDS
        while int(time.time()) < time_out_timestamp:
            if authentication_id in cls.store:
                return cls.store.pop(authentication_id)

            await asyncio.sleep(constants.ACQUIRE_ACCESS_TOKEN__POLLING_SLEEP_SECONDS)

        raise AccessTokenRetrievalTimeoutException(f"Unauthenticated authentication: {authentication_id}")


class MicrosoftIdentity(DataClass):
    access_token: str
    audience_id: str
    authenticated_timestamp: datetime.datetime
    inception_timestamp: datetime.datetime
    expiry_timestamp: datetime.datetime
    application_id: str
    name: str
    scope: str
    user_id: str

    @classmethod
    def get_identity_from_access_token(cls, access_token: str) -> "MicrosoftIdentity":
        try:
            payload: Dict[str, Any] = jwt.decode(access_token, options={"verify_signature": False})
            return MicrosoftIdentity(
                access_token=access_token,
                audience_id=payload["aud"],
                authenticated_timestamp=datetime.datetime.utcfromtimestamp(payload["iat"]),
                inception_timestamp=datetime.datetime.utcfromtimestamp(payload["nbf"]),
                expiry_timestamp=datetime.datetime.utcfromtimestamp(payload["exp"]),
                application_id=payload["appid"],
                name=f"{payload['given_name']} {payload['family_name']}",
                scope=payload["scp"],
                user_id=payload["unique_name"]
            )

        except Exception:
            raise InvalidAccessTokenException("Invalid token")

    @staticmethod
    def get_login_url(redirect_url: str,
                      authentication_id: str,
                      entra_id_tenant_id: str | None = None,
                      entra_id_client_id: str | None = None,
                      scope: str | None = None,
                      url_shortener_function: Union[Callable, None] = None) -> str:

        entra_id_tenant_id = common.get_environmental_secret(EnvironmentSecret.ENTRA_ID_TENANT_ID) if entra_id_tenant_id is None else entra_id_tenant_id
        entra_id_client_id = common.get_environmental_secret(EnvironmentSecret.ENTRA_ID_CLIENT_ID) if entra_id_client_id is None else entra_id_client_id
        scope = "User.Read" if scope is None else scope

        params: Dict[str, str] = {"client_id": entra_id_client_id,
                                  "response_type": "code",
                                  "redirect_uri": redirect_url,
                                  "response_mode": "query",
                                  "scope": scope,
                                  "state": authentication_id,
                                  "code_challenge_method": "S256",
                                  "code_challenge": base64.urlsafe_b64encode(hashlib.sha256(authentication_id.encode('utf-8')).digest()).decode('utf-8').replace("=", "")}
        url: str = f"https://login.microsoftonline.com/{entra_id_tenant_id}/oauth2/v2.0/authorize?{urlencode(params)}"

        return url if url_shortener_function is None else url_shortener_function(url)

    @staticmethod
    async def get_access_token_remotely(redirect_url: str, discord_text_channel_name: str | None = None, url_shortener_function: Union[Callable, None] = None) -> str:
        authentication_id: str = common.get_unique_id()
        discord_text_channel_name = AortaTextChannels.NOTIFICATION.value if discord_text_channel_name is None else discord_text_channel_name
        sign_in_url: str = MicrosoftIdentity.get_login_url(redirect_url, authentication_id) if url_shortener_function is None else url_shortener_function(MicrosoftIdentity.get_login_url(redirect_url, authentication_id))

        await DiscordDefaults.send_message(discord_text_channel_name, "**Authentication Request**\n"
                                                                      f"*Sign-in here*: {sign_in_url}")

        authentication_code: str = await MicrosoftEntraIDAuthenticationIDStore._get_or_wait(authentication_id)
        return await MicrosoftIdentity._get_access_token_from_authentication_code(authentication_code, authentication_id, redirect_url)

    @staticmethod
    async def _get_access_token_from_authentication_code(authentication_code: str, authentication_id: str, redirect_url: str, entra_id_tenant_id: str | None = None, entra_id_client_id: str | None = None) -> str:
        entra_id_tenant_id = common.get_environmental_secret(EnvironmentSecret.ENTRA_ID_TENANT_ID) if entra_id_tenant_id is None else entra_id_tenant_id
        entra_id_client_id = common.get_environmental_secret(EnvironmentSecret.ENTRA_ID_CLIENT_ID) if entra_id_client_id is None else entra_id_client_id
        url: str = f"https://login.microsoftonline.com/{entra_id_tenant_id}/oauth2/v2.0/token"

        try:
            response: HTTPResponse = await AsyncHTTPSession(url).post(url, data={"client_id": entra_id_client_id,
                                                                                 "redirect_uri": redirect_url,
                                                                                 "code": authentication_code,
                                                                                 "grant_type": "authorization_code",
                                                                                 "code_verifier": authentication_id}, is_form_url_encoded=True)
            return response.data["access_token"]
        except ClientSideException as e:
            response = e.data["http_response"]
            raise ClientSideException(response.data["error_description"])



import datetime

from sirius.communication import discord
from sirius.communication.discord import TextChannel, AortaTextChannels, DiscordDefaults


class Logger:

    @staticmethod
    async def _send_message(text_channel_enum: AortaTextChannels, message: str) -> None:
        await DiscordDefaults.send_message(text_channel_enum.value, f"{discord.get_timestamp_string(datetime.datetime.now())}: {message}")

    @staticmethod
    async def notify(message: str) -> None:
        await Logger._send_message(AortaTextChannels.NOTIFICATION, message)

    @staticmethod
    async def debug(message: str) -> None:
        await Logger._send_message(AortaTextChannels.DEBUG, message)



from typing import Dict, Any


class SiriusException(Exception):
    data: Dict[str, Any] = {}

    def __init__(self, message: str, data: Dict[str, Any] | None = None) -> None:
        super().__init__(message)
        self.data = data if data is not None else self.data


class ApplicationException(SiriusException):
    pass


class SDKClientException(SiriusException):
    pass


class OperationNotSupportedException(SDKClientException):
    pass



import asyncio
import datetime
import time
from enum import Enum
from http import HTTPStatus
from logging import Logger
from typing import List, Dict, Any, Union

from pydantic import PrivateAttr

from sirius import application_performance_monitoring, common
from sirius.common import DataClass
from sirius.communication.discord import constants
from sirius.communication.discord.exceptions import ServerNotFoundException, DuplicateServersFoundException, RoleNotFoundException
from sirius.constants import EnvironmentSecret
from sirius.exceptions import OperationNotSupportedException
from sirius.http_requests import AsyncHTTPSession, HTTPResponse, ClientSideException

logger: Logger = application_performance_monitoring.get_logger()
default_bot: Union["Bot", None] = None


class ServerName(Enum):
    VITA: str = "Vita"
    AURUM: str = "Aurum"
    AORTA: str = "Aorta"


class AortaTextChannels(Enum):
    DEBUG: str = "logs"
    NOTIFICATION: str = "notification"
    WISE: str = "wise"


class RoleType(Enum):
    EVERYONE: str = "@everyone"
    BOT: str = "Bot"
    OTHER: str = ""


class DiscordHTTPSession(AsyncHTTPSession):

    async def get(self, url: str, query_params: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        try:
            return await super().get(url, query_params, headers)
        except ClientSideException as e:
            http_response: HTTPResponse = e.data["http_response"]

            if http_response.response_code == HTTPStatus.TOO_MANY_REQUESTS:
                await asyncio.sleep(http_response.data["retry_after"] + 0.1)
                return await self.get(url, query_params, headers)

            raise e

    async def put(self, url: str, data: Dict[str, Any], headers: Dict[str, Any] | None = None) -> HTTPResponse:
        try:
            return await super().put(url, data, headers)
        except ClientSideException as e:
            http_response: HTTPResponse = e.data["http_response"]

            if http_response.response_code == HTTPStatus.TOO_MANY_REQUESTS:
                await asyncio.sleep(http_response.data["retry_after"] + 0.1)
                return await self.put(url, data, headers)

            raise e

    async def post(self, url: str, data: Dict[str, Any] | None = None, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        try:
            return await super().post(url, data, headers)
        except ClientSideException as e:
            http_response: HTTPResponse = e.data["http_response"]

            if http_response.response_code == HTTPStatus.TOO_MANY_REQUESTS:
                await asyncio.sleep(http_response.data["retry_after"] + 0.1)
                return await self.post(url, data, headers)

            raise e

    async def delete(self, url: str, headers: Dict[str, Any] | None = None) -> HTTPResponse:
        try:
            return await super().delete(url, headers)
        except ClientSideException as e:
            http_response: HTTPResponse = e.data["http_response"]

            if http_response.response_code == HTTPStatus.TOO_MANY_REQUESTS:
                await asyncio.sleep(http_response.data["retry_after"] + 0.1)
                return await self.delete(url, headers)

            raise e


DiscordHTTPSession(constants.URL,
                   {"Authorization": f"Bot {common.get_environmental_secret(EnvironmentSecret.DISCORD_BOT_TOKEN)}"})


class DiscordDefaults(DataClass):

    @staticmethod
    async def send_message(text_channel_name: str, message: str) -> None:
        global default_bot
        default_bot = await Bot.get() if default_bot is None else default_bot
        server: Server = await default_bot.get_server()
        text_channel: TextChannel = await server.get_text_channel(text_channel_name)
        await text_channel.send_message(message)

    @classmethod
    async def notify(cls, message: str) -> None:
        await DiscordDefaults.send_message(AortaTextChannels.NOTIFICATION.value, message)


class Bot(DataClass):
    id: int
    username: str
    name: str
    server_list: List["Server"] = []
    _http_session: DiscordHTTPSession = PrivateAttr()

    @property
    def http_session(self) -> DiscordHTTPSession:
        return self._http_session

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    async def _initialize(self) -> None:
        for server in await Server.get_all_servers(self):
            if server in self.server_list:
                existing_server: Server = next(filter(lambda s: s.id == server.id, self.server_list))
                existing_server.__dict__.update(**server.model_dump())
            else:
                self.server_list.append(server)

    async def get_server(self, server_name: str | None = None) -> "Server":
        if server_name is None:
            server_name = Server.get_default_server_name()

        server_list: List[Server] = list(filter(lambda s: s.name == server_name, self.server_list))
        if len(server_list) == 1:
            return server_list[0]
        elif len(server_list) == 0:
            raise ServerNotFoundException(f"Server not found\n"
                                          f"Bot Name: {self.username}\n"
                                          f"Server Name: {server_name}\n")
        else:
            raise DuplicateServersFoundException(f"Duplicate servers found\n"
                                                 f"Bot Name: {self.username}\n"
                                                 f"Server Name: {server_name}\n"
                                                 f"Number of servers: {len(server_list)}\n")

    @staticmethod
    async def get() -> "Bot":
        url: str = constants.ENDPOINT__BOT__GET_BOT
        http_session: DiscordHTTPSession = DiscordHTTPSession(url)
        response: HTTPResponse = await http_session.get(url)

        bot: Bot = Bot.model_construct(id=response.data["id"],
                                       username=response.data["username"],
                                       name=response.data["global_name"])
        bot._http_session = http_session
        await bot._initialize()

        return bot


class Server(DataClass):
    id: int
    name: str
    text_channel_list: List["TextChannel"] = []
    user_list: List["User"] = []
    role_list: List["Role"] = []
    bot: Bot

    @property
    def http_session(self) -> DiscordHTTPSession:
        return self.bot.http_session

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    async def _initialize(self) -> None:
        for text_channel in await TextChannel.get_all(self):
            if text_channel in self.text_channel_list:
                existing_text_channel: TextChannel = next(filter(lambda t: t.id == text_channel.id, self.text_channel_list))
                existing_text_channel.__dict__.update(**text_channel.model_dump())
            else:
                self.text_channel_list.append(text_channel)

        for user in await User.get_all(self):
            if user in self.user_list:
                existing_user: User = next(filter(lambda u: u.id == user.id, self.user_list))
                existing_user.__dict__.update(user.model_dump())
            else:
                self.user_list.append(user)

        for role in await Role.get_all(self):
            if role in self.role_list:
                existing_role: Role = next(filter(lambda r: r.id == role.id, self.role_list))
                existing_role.__dict__.update(role.model_dump())
            else:
                self.role_list.append(role)

    async def get_text_channel(self, text_channel_name: str, is_public_channel: bool = False) -> "TextChannel":
        text_channel_list: List[TextChannel] = list(filter(lambda t: t.name == text_channel_name, self.text_channel_list))

        if len(text_channel_list) == 1:
            return text_channel_list[0]
        elif len(text_channel_list) == 0:
            logger.warning("Channel not found; creating channel\n"
                           f"Server Name: {self.name}\n"
                           f"Channel Name: {text_channel_name}\n"
                           )
            text_channel: TextChannel = await TextChannel.create(text_channel_name, self, is_public_channel=is_public_channel)
            self.text_channel_list.append(text_channel)
            return text_channel
        else:
            raise DuplicateServersFoundException(f"Duplicate channels found\n"
                                                 f"Server Name: {self.name}\n"
                                                 f"Channel Name: {text_channel_name}\n"
                                                 f"Number of channels: {len(text_channel_list)}\n")

    async def get_user(self, username: str) -> "User":
        username = username.lower()
        self.user_list = await User.get_all(self) if len(self.user_list) == 0 else self.user_list

        try:
            return next(filter(lambda u: u.username == username, self.user_list))
        except StopIteration:
            raise RoleNotFoundException(f"User not found\n"
                                        f"Server Name: {self.name}\n"
                                        f"Username: {username}")

    async def get_role(self, role_type: RoleType) -> "Role":
        if role_type == RoleType.OTHER:
            raise OperationNotSupportedException("OTHER Role Type searches are not allowed")

        self.role_list = await Role.get_all(self) if len(self.role_list) == 0 else self.role_list

        try:
            return next(filter(lambda r: r.role_type == role_type, self.role_list))
        except StopIteration:
            raise RoleNotFoundException(f"User not found\n"
                                        f"Server Name: {self.name}\n"
                                        f"Role Type: {role_type.value}")

    @classmethod
    async def get_all_servers(cls, bot: Bot) -> List["Server"]:
        server_list: List[Server] = []
        response: HTTPResponse = await bot.http_session.get(constants.ENDPOINT__SERVER__GET_ALL_SERVERS)

        for data in response.data:
            server: Server = Server.model_construct(id=data["id"],
                                                    name=data["name"])
            server.bot = bot
            await server._initialize()
            server_list.append(server)

        return server_list

    @staticmethod
    def get_default_server_name() -> str:
        server_name: str = common.get_application_name()
        return server_name if common.is_production_environment() else f"{server_name} [Dev]"


class User(DataClass):
    id: int
    username: str
    name: str | None
    is_bot: bool
    server: Server

    @property
    def http_session(self) -> DiscordHTTPSession:
        return self.server.bot.http_session

    @staticmethod
    async def get_all(server: Server) -> List["User"]:
        url: str = constants.ENDPOINT__SERVER__GET_ALL_USERS.replace("$serverID", str(server.id))
        http_session: DiscordHTTPSession = DiscordHTTPSession(url)
        response: HTTPResponse = await http_session.get(url)
        return [User(id=data["user"]["id"],
                     username=data["user"]["username"],
                     name=data["user"]["global_name"],
                     is_bot=data["user"]["bot"] if "bot" in data["user"] else False,
                     server=server) for data in response.data]


class Role(DataClass):
    id: int
    role_type: RoleType
    permissions: str
    server: Server

    @property
    def http_session(self) -> DiscordHTTPSession:
        return self.server.bot.http_session

    @staticmethod
    async def get_all(server: Server) -> List["Role"]:
        role_list: List[Role] = []
        url: str = constants.ENDPOINT__SERVER__GET_ALL_ROLES.replace("$serverID", str(server.id))
        http_session: DiscordHTTPSession = DiscordHTTPSession(url)
        response: HTTPResponse = await http_session.get(url)

        for data in response.data:
            try:
                role_type: RoleType = RoleType(data["name"])
            except ValueError:
                role_type = RoleType.OTHER

            role_list.append(Role(id=data["id"],
                                  role_type=role_type,
                                  permissions=data["permissions"],
                                  server=server))
        return role_list


class Channel(DataClass):
    id: int
    name: str
    type: int
    server: Server

    @property
    def http_session(self) -> DiscordHTTPSession:
        return self.server.bot.http_session

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    @classmethod
    async def get_all_channels(cls, server: Server) -> List["Channel"]:
        url: str = constants.ENDPOINT__CHANNEL__CREATE_CHANNEL_OR_GET_ALL_CHANNELS.replace("$serverID", str(server.id))
        response: HTTPResponse = await server.http_session.get(url)

        return [Channel(id=data["id"],
                        name=data["name"],
                        type=data["type"],
                        server=server) for data in response.data]

    @classmethod
    async def create(cls, channel_name: str, server: Server, type_id: int, is_public_channel: bool = False) -> "Channel":
        url: str = constants.ENDPOINT__CHANNEL__CREATE_CHANNEL_OR_GET_ALL_CHANNELS.replace("$serverID", str(server.id))
        data: Dict[str, Any] = {"name": channel_name, "type": type_id}

        if not is_public_channel:
            data["permission_overwrites"] = [{
                "id": str((await server.get_role(RoleType.EVERYONE)).id),
                "type": 0,
                "allow": 0,
                "deny": 1024
            }]

        response: HTTPResponse = await server.http_session.post(url, data=data)
        return Channel(id=response.data["id"],
                       name=response.data["name"],
                       type=response.data["type"],
                       server=server)


class TextChannel(Channel):

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

    #   TODO: This is no longer an asynchronous function
    async def send_message(self, message: str) -> None:
        if common.is_ci_cd_pipeline_environment():
            return

        url: str = constants.ENDPOINT__CHANNEL__SEND_MESSAGE.replace("$channelID", str(self.id))
        asyncio.ensure_future(self.http_session.post(url, data={"content": message}))

    async def delete(self) -> None:
        await self.http_session.delete(constants.ENDPOINT__CHANNEL__DELETE.replace("$channelID", str(self.id)))
        self.server.text_channel_list.remove(self)

    @classmethod
    async def get_all(cls, server: Server) -> List["TextChannel"]:
        text_channel_list: List[TextChannel] = []
        channel_list: List[Channel] = list(filter(lambda c: c.type == 0, await Channel.get_all_channels(server)))

        for channel in channel_list:
            text_channel: TextChannel = TextChannel.model_construct(**channel.model_dump(exclude={"server"}))
            text_channel.server = server
            text_channel_list.append(text_channel)

        return text_channel_list

    @classmethod
    async def create(cls, text_channel_name: str, server: Server, type_id: int = 0, is_public_channel: bool = False) -> "TextChannel":
        channel: Channel = await Channel.create(text_channel_name, server, type_id, is_public_channel)
        text_channel: TextChannel = TextChannel.model_construct(**channel.model_dump(exclude={"server"}))
        text_channel.server = server
        return text_channel


class Message(DataClass):
    id: int
    content: str


def get_timestamp_string(timestamp: datetime.datetime | datetime.date) -> str:
    return f"<t:{str(int(time.mktime(timestamp.timetuple())))}:T>"
